#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define USE_CONSOLE
#include <allegro.h>

#define FALSE 0
#define TRUE -1

typedef struct
{
   long version;
   long unknown1;
   long unknown2;
   int  termination[4];
   long unknown3;
   long blockcount;
} DC6HEADER;

typedef struct
{
	long unknown1;
	long width;
	long height;
	long offset_x;
	long offset_y;
	long unknown2;
	long next_block;
	long length;
} DC6BLOCKHEADER;

typedef struct
{
   char           filename[257];
   DC6HEADER      dc6header;
   DC6BLOCKHEADER * dc6blockheader;
   long           * block_ptr;
   BITMAP         * bmp;
} HEADERS_ELEMENTS;

typedef struct
{
   char * palette;
   char * background;
   int    verbose;
   int    skill;
   int    blur;
   int    from;
   int    to;
} PARAMS;

typedef struct
{
   BITMAP * bmp;
   int    width;
   int    height;
   int    min_x;
   int    min_y;
   int    max_x;
   int    max_y;
} BOX;

PALETTE dc6pal;


HEADERS_ELEMENTS body, skill;
PARAMS           glb_prm;
BOX              box;


// ==========================================================================
int read_dc6_header(FILE * in, HEADERS_ELEMENTS * elm)
{
   DC6HEADER * ptr = & elm->dc6header;
   long      * lptr;
   int       i, nb;
   
   fread(&ptr->version,        4, 1, in);
   fread(&ptr->unknown1,       4, 1, in);
   fread(&ptr->unknown2,       4, 1, in);
   fread(&ptr->termination[0], 1, 1, in);
   fread(&ptr->termination[1], 1, 1, in);
   fread(&ptr->termination[2], 1, 1, in);
   fread(&ptr->termination[3], 1, 1, in);
   fread(&ptr->unknown3,       4, 1, in);
   fread(&ptr->blockcount,     4, 1, in);

   if (glb_prm.verbose)
   {
      printf("version     : %li\n", ptr->version);
      printf("unknown1    : %li\n", ptr->unknown1);
      printf("unknown2    : %li\n", ptr->unknown2);
      printf("termination : %02X %02X %02X %02X\n",
         ptr->termination[0],
         ptr->termination[1],
         ptr->termination[2],
         ptr->termination[3]);
      printf("unknown3    : %li\n",   ptr->unknown3);
      printf("blockcount  : %li\n\n", ptr->blockcount);
   }

   nb = ptr->blockcount;
   
   elm->block_ptr = (long *) malloc(sizeof(long) * nb);
   if (elm->block_ptr == NULL)
   {
      printf("not enough memory for %li block pointers\n", nb);
      return 1;
   }
   
   elm->dc6blockheader = (DC6BLOCKHEADER *) malloc(sizeof(DC6BLOCKHEADER) * nb);
   if (elm->dc6blockheader == NULL)
   {
      printf("not enough memory for %li block headers\n", nb);
      return 1;
   }
   
   lptr = elm->block_ptr;
   for (i=0; i<nb; i++)
   {
      fread(lptr, 4, 1, in);
      if (glb_prm.verbose)
         printf("   block pointer %3i : %li\n", i, * lptr);
      lptr++;
   }
   return 0;
}

// ==========================================================================
void read_dc6_blockheader(FILE * in, int i, HEADERS_ELEMENTS * elm)
{
   DC6BLOCKHEADER * bh = elm->dc6blockheader + i, * bh_old;
   long tmp_off, hole;
   int  x;
   long old_seek;
   
   fseek(in, * (elm->block_ptr + i), SEEK_SET);

   fread(&bh->unknown1,   4, 1, in);
   fread(&bh->width,      4, 1, in);
   fread(&bh->height,     4, 1, in);
   fread(&bh->offset_x,   4, 1, in);
   fread(&bh->offset_y,   4, 1, in);
   fread(&bh->unknown2,   4, 1, in);
   fread(&bh->next_block, 4, 1, in);
   fread(&bh->length,     4, 1, in);
   
   if (i > 0)
   {
      tmp_off =  * (elm->block_ptr + i - 1);
      tmp_off += 4 * 8;
      bh_old  =  elm->dc6blockheader + i - 1;
      tmp_off += bh_old->length;
      hole    =  * (elm->block_ptr + i) - tmp_off;
      if (hole && glb_prm.verbose)
      {
         printf("\n      HOLE       : %li ( ", hole);
         old_seek = ftell(in);
         fseek(in, * (elm->block_ptr + i) - hole, SEEK_SET);
         for (x=0; x<hole; x++)
            printf("%02X ", fgetc(in));
         printf(")\n");
         fseek(in, old_seek, SEEK_SET);
      }
   }
   if (glb_prm.verbose)
   {
      printf("\n   block header %03i (offset %li in file):\n", i, * (elm->block_ptr+i));
      printf("      unknown1   : %li\n", bh->unknown1);
      printf("      width      : %li\n", bh->width);
      printf("      height     : %li\n", bh->height);
      printf("      offset_x   : %li\n", bh->offset_x);
      printf("      offset_y   : %li\n", bh->offset_y);
      printf("      unknown2   : %li\n", bh->unknown2);
      printf("      next_block : %li\n", bh->next_block);
      printf("      length     : %li\n", bh->length);
   }
}

// ==========================================================================
void search_box(void)
{
   DC6BLOCKHEADER * bh, * bh_ski;
   long           nb, i, min_x, max_x, min_y, max_y, x1, y1, x2, y2;

   min_x = 2000000L;
   max_x = -2000000L;
   min_y = 2000000L;
   max_y = -2000000L;

   nb = body.dc6header.blockcount;
   if (glb_prm.verbose)
      printf("\nanalyse dc6 \"body\" file\n");
   bh = body.dc6blockheader;
   for (i=0; i<nb; i++)
   {
      if (glb_prm.verbose)
         printf("frame %3li : %3li, %3li (%4li, %4li) ---> ",
            i, bh->width, bh->height, bh->offset_x, bh->offset_y);
      x1 = bh->offset_x;
      y1 = bh->offset_y;
      x2 = x1 + bh->width;
      y2 = y1 - bh->height;
      
      if (x1<min_x) min_x = x1;
      if (x2>max_x) max_x = x2;
      if (y1<min_y) min_y = y1;
      if (y2>max_y) max_y = y2;
      
      if (x2<min_x) min_x = x2;
      if (x1>max_x) max_x = x1;
      if (y2<min_y) min_y = y2;
      if (y1>max_y) max_y = y1;
      
      if (glb_prm.verbose)
         printf("(%4li, %4li) TO (%4li, %4li)\n", x1, y1, x2, y2);
      bh++;
   }

   if (glb_prm.skill)
   {
      nb = skill.dc6header.blockcount;
      if (glb_prm.verbose)
         printf("\nanalyse dc6 \"skill\" file\n");
      bh = body.dc6blockheader;
      bh_ski = skill.dc6blockheader;
      for (i=0; i<nb; i++)
      {
         if (glb_prm.verbose)
            printf("frame %3li : %3li, %3li (%4li, %4li) ---> ",
               i, bh_ski->width, bh_ski->height, bh_ski->offset_x, bh_ski->offset_y);

//      off_x = bod_bh->offset_x - box.min_x;
//      off_y = bod_bh->offset_y - box.min_y;

//      off_x = ski_bh->offset_x - box.min_x ;
//      off_y = bod_bh->height + ski_bh->offset_y - ski_bh->height;

         x1 = bh_ski->offset_x;
         y1 = bh_ski->offset_y;
//         y1 = bh->height + bh_ski->offset_y - bh_ski->height;
         x2 = x1 + bh_ski->width;
         y2 = y1 - bh_ski->height;

         if (x1<min_x) min_x = x1;
         if (x2>max_x) max_x = x2;
         if (y1<min_y) min_y = y1;
         if (y2>max_y) max_y = y2;
      
         if (x2<min_x) min_x = x2;
         if (x1>max_x) max_x = x1;
         if (y2<min_y) min_y = y2;
         if (y1>max_y) max_y = y1;
      
         if (glb_prm.verbose)
            printf("(%4li, %4li) TO (%4li, %4li)\n", x1, y1, x2, y2);
         bh++;
         bh_ski++;
      }
   }

   box.width  = max_x - min_x + 1;
   box.height = max_y - min_y + 1;
   box.min_x  = min_x;
   box.min_y  = min_y;
   box.max_x  = max_x;
   box.max_y  = max_y;
   if (glb_prm.verbose)
   {
      printf("\nRESULT :\n");
      printf("min x, y = %4li, %4li\n", min_x, min_y);
      printf("max x, y = %4li, %4li\n", max_x, max_y);
      printf(" ==> box = %4li, %4li\n", box.width, box.height);
   }
}

// ==========================================================================
int analyse_commandline(int argc, char * argv[])
{
   int i, help = FALSE;
   
   // help
   for (i=1; i<argc && (stricmp("-h", argv[i]) != 0) &&
                       (strstr("?", argv[i]) == NULL) ; i++)
      ;
   if ((i < argc) || (argc == 1))
   {
      printf("\nsyntaxe : dc6extr <file.dc6> [options]\n"
             "   options are :\n"
             "      -?           : this help\n"
             "      -v           : verbose mode\n"
             "      -pal <file>  : use a diablo 2 palette\n\n"
             "      -skill       : check the skills frame to make the body frames\n"
             "                     ex : \"dc6extr nenu3.dc6 -skill\" will extract frames of\n"
             "                     nenu3.dc6 and nenu3s.dc6, same size for all bitmaps.\n"
             "                     Skill's dc6 need to be in the same place of body's dc6\n\n"
             "      -blur        : default method is : raw blit of the skill frames over the\n"
             "                     body frames. With -blur, light them with the skill frames\n"
             "                        . use -blur with the necro and sorc dc6 files\n"
             "                        . for amazon, paladin and so on : don't use it\n\n"
             "      -back <file> : use a file as a background (instead of a pink color)\n"
             "                     usefull if you need to make a web animated gif file\n"
             "                     need to be the exact size of the resulting bitmap files\n\n"
             "      -from <x>    : extract from the x frame (first is 0) \n"
             "      -to   <y>    : extract up to the y frame\n"
      );
      return 1;
   }
   
   // verbose
   for (i=2; i<argc && stricmp("-v", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.verbose = TRUE;

   // palette
   for (i=2; i<argc && stricmp("-pal", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.palette = argv[i+1];

   // skill
   for (i=2; i<argc && stricmp("-skill", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.skill = TRUE;

   // blur
   for (i=2; i<argc && stricmp("-blur", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.blur = TRUE;

   // background
   for (i=2; i<argc && stricmp("-back", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.background = argv[i+1];

   // starting frame
   for (i=2; i<argc && stricmp("-from", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.from = atoi(argv[i+1]);

   // ending frame
   for (i=2; i<argc && stricmp("-to", argv[i]) != 0; i++)
      ;
   if (i < argc)
      glb_prm.to = atoi(argv[i+1]);

   return 0;
}


// ==========================================================================
int decompress_dc6(FILE * in, int frame, HEADERS_ELEMENTS * elm)
{
   DC6BLOCKHEADER * bh;
   long           i, i2;
   int            c, c2, x, y, color;
   
   bh = elm->dc6blockheader + frame;
   elm->bmp = create_bitmap(bh->width, bh->height);
   clear_to_color(elm->bmp, makeacol32(255, 0, 255, 0));
   fseek(in, (* (elm->block_ptr + frame)) + 32, SEEK_SET);
   x = 0;
   y = bh->height - 1;

   for (i=0; i<bh->length; i++)
   {
      c = fgetc(in);
      if (c == 0x80)
      {
         x = 0;
         y--;
      }
      else if (c & 0x80)
         x += c & 0x7F;
      else
      {
         for (i2=0; i2<c; i2++)
         {
            c2 = fgetc(in);
            color = makeacol32(dc6pal[c2].r, dc6pal[c2].g, dc6pal[c2].b, 255);
            putpixel(elm->bmp, x, y, color);
            x++;
         }
      }
   }

   if (elm == &body)
      blit(elm->bmp, screen, 0, 0, 10, 10, bh->width, bh->height);
   else if (elm == &skill)
      blit(elm->bmp, screen, 0, 0, 200, 10, bh->width, bh->height);
   return 0;
}


// ==========================================================================
void make_frame(int frame)
{
   DC6BLOCKHEADER * bod_bh, * ski_bh;
   int            off_x, off_y;
   char           str_tmp[256], fmt[] = "out-%03i.tga";
   int            x, y, s, d, sr, sg ,sb ,dr ,dg ,db, fr, fg, fb, pink;
   
   set_alpha_blender();
//   clear_to_color(box.bmp, makeacol32(255, 0, 255, 0)); // PINK
   clear_to_color(box.bmp, makeacol32(0, 0, 0, 0)); // BLACK
   
   if (frame < body.dc6header.blockcount)
   {
      bod_bh = body.dc6blockheader  + frame;
      off_x = bod_bh->offset_x - box.min_x;
      off_y = bod_bh->offset_y - box.min_y - bod_bh->height;
      draw_trans_sprite(box.bmp, body.bmp, off_x, off_y);
   }
   
   if (frame < skill.dc6header.blockcount)
   {
      ski_bh = skill.dc6blockheader + frame;
      off_x = ski_bh->offset_x - box.min_x ;
      off_y = ski_bh->offset_y - box.min_y - ski_bh->height;

      if (glb_prm.blur)
      {
printf("BLUR\n");
         pink = makecol32(255, 0, 255);
         for (y=0; y<skill.bmp->h; y++)
         {
            for (x=0; x<skill.bmp->w; x++)
            {
               s = getpixel(skill.bmp, x, y);
               if (s != pink)
               {
                  sr = getr(s);
                  sg = getg(s);
                  sb = getb(s);
                  d = getpixel(box.bmp, off_x + x, off_y + y);
                  dr = getr(d);
                  dg = getg(d);
                  db = getb(d);
                  fr = dr + sr <= 255 ? dr + sr : 255;
                  fg = dg + sg <= 255 ? dg + sg : 255;
                  fb = db + sb <= 255 ? db + sb : 255;
                  d = makeacol(fr, fg, fb, 255);
                  putpixel(box.bmp, off_x + x, off_y + y, d);
               }
            }
         }
      }
      else
         draw_trans_sprite(box.bmp, skill.bmp, off_x, off_y);
   }
   
   vsync();
   blit(box.bmp, screen, 0, 0, 400, 10, box.width, box.height);
   sprintf(str_tmp, fmt, frame);
   save_tga(str_tmp, box.bmp, NULL);
}


// ==========================================================================
int load_palette(void)
{
   FILE * in;
   int  i;
   
   if (glb_prm.palette != NULL)
   {
      in = fopen(glb_prm.palette, "rb");
      if (in == NULL)
      {
         printf("can't open palette : %s\n", glb_prm.palette);
         return 1;
      }
      printf("opening palette : %s\n", glb_prm.palette);
      for (i=0; i<256; i++)
      {
         dc6pal[i].b = fgetc(in);
         dc6pal[i].g = fgetc(in);
         dc6pal[i].r = fgetc(in);
      }
   }
   fclose(in);
   return 0;
}

// ==========================================================================
int main(int argc, char * argv[])
{
   FILE * body_f, * skill_f = NULL;
   int  i, f, t;
   char tmp_str[257];

   memset(&body,  0, sizeof (body));
   memset(&skill, 0, sizeof (skill));
   memset(&glb_prm, 0, sizeof (glb_prm));
   glb_prm.to = -1;

   if (0 != analyse_commandline(argc, argv))
      return 1;

   if (argc > 2)
   {
      for (i=2; i<argc && stricmp("-s", argv[i]) != 0; i++)
         ;
      if (i < argc)
         glb_prm.skill = TRUE;
   }
   
   // regular file
   body_f = fopen(argv[1], "rb");
   if (body_f == NULL)
   {
      printf("%s not found\n", argv[1]);
      return 1;
   }
   strcpy(body.filename, argv[1]);
   if (glb_prm.verbose)
   {
      printf("====================\n%s\n====================\n", argv[1]);
      printf("file size   : ");
      fseek(body_f, 0, SEEK_END);
      printf("%li\n\n", ftell(body_f));
      fseek(body_f, 0, SEEK_SET);
   }
   
   if (read_dc6_header(body_f, &body) == 0)
   {
      for (i=0; i<body.dc6header.blockcount; i++)
         read_dc6_blockheader(body_f, i, &body);
   }

   // skill file
   if (glb_prm.skill)
   {
      if (glb_prm.verbose)
         printf("\n\n******* check for dc6 skill file *******\n\n");
      strcpy(tmp_str, argv[1]);
      tmp_str[strlen(tmp_str)-4] = 0;
      strcat(tmp_str, "s.dc6");
      skill_f = fopen(tmp_str, "rb");
      if (skill_f == NULL)
      {
         printf("%s not found\n", tmp_str);
         glb_prm.skill = FALSE;
      }
      else
      {
         strcpy(skill.filename, tmp_str);
         if (glb_prm.verbose)
         {
            printf("====================\n%s\n====================\n", tmp_str);
            printf("file size   : ");
            fseek(skill_f, 0, SEEK_END);
            printf("%li\n\n", ftell(skill_f));
            fseek(skill_f, 0, SEEK_SET);
         }
         if (read_dc6_header(skill_f, &skill) == 0)
         {
            for (i=0; i<skill.dc6header.blockcount; i++)
               read_dc6_blockheader(skill_f, i, &skill);
         }
      }
   }

   // extract bitmaps
   search_box();
   allegro_init();
   install_keyboard();

   set_color_depth(32);
   if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0))
   {
      set_color_depth(24);
      if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0))
      {
         printf("can't select a 640 x 480 truecolor graphic mode (24 or 32 bits depth)\n");
         fclose(body_f);
         fclose(skill_f);
         free(body.block_ptr);
         free(body.dc6blockheader);
         free(skill.block_ptr);
         free(skill.dc6blockheader);
         return 1;
      }
   }

   if (load_palette() == 0)
      set_palette(dc6pal);

   box.bmp = create_bitmap(box.width, box.height);
   f = glb_prm.from;
   if (f < 0)
      f = 0;
   t = glb_prm.to != -1 ? glb_prm.to + 1 : body.dc6header.blockcount;
   if (t > body.dc6header.blockcount) t = body.dc6header.blockcount;
   if (t <= f) t = f + 1;

//readkey();

clear_to_color(screen, makecol(50, 0, 50));
   for (i=f; i<t; i++)
   {
      decompress_dc6(body_f, i, &body);
      if (glb_prm.skill)
         decompress_dc6(skill_f, i, &skill);
      make_frame(i);

      destroy_bitmap(body.bmp);
      if (glb_prm.skill)
         destroy_bitmap(skill.bmp);
   }
   text_mode(-1);
   textout(screen, font, "PRESS A KEY TO QUIT / APPUYEZ SUR UNE TOUCHE POUR QUITTER",
      10, 460, makecol(255, 255, 255));
readkey();

   // close
printf("close body_f ...");
fflush(stdout);
   fclose(body_f);
printf("ok\n");
fflush(stdout);

printf("close skill_f ...");
fflush(stdout);
   fclose(skill_f);
printf("ok\n");
fflush(stdout);
   
printf("destroy_bitmap box ...");
fflush(stdout);
   destroy_bitmap(box.bmp);
printf("ok\n");
fflush(stdout);

printf("free.block_ptr ...");
fflush(stdout);
   free(body.block_ptr);
printf("ok\n");
fflush(stdout);

printf("free body.dc6blockheader ...");
fflush(stdout);
   free(body.dc6blockheader);
printf("ok\n");
fflush(stdout);

printf("free skill.block_ptr ...");
fflush(stdout);
   free(skill.block_ptr);
printf("ok\n");
fflush(stdout);

printf("free skill.dc6blockheader ...");
fflush(stdout);
   free(skill.dc6blockheader);
printf("ok\n");
fflush(stdout);

   return 0;
}
END_OF_MAIN();
